@@ -15,10 +15,10 @@
 
 //! Functions and data structures of the swayr client.
 
-use crate::con;
-use crate::con::NodeMethods;
 use crate::config as cfg;
 use crate::layout;
+use crate::tree as t;
+use crate::tree::NodeMethods;
 use crate::util;
 use crate::util::DisplayFormat;
 use lazy_static::lazy_static;
@@ -52,6 +52,20 @@ pub enum SwayrCommand {
     SwitchToUrgentOrLRUWindow,
     /// Focus the selected window.
     SwitchWindow,
+    /// Switch to the selected workspace.
+    SwitchWorkspace,
+    /// Switch to the selected workspace or focus the selected window.
+    SwitchWorkspaceOrWindow,
+    /// Switch to the selected workspace or focus the selected container, or
+    /// window.
+    SwitchWorkspaceContainerOrWindow,
+    /// Quit the selected window.
+    QuitWindow,
+    /// Quit all windows of selected workspace or the selected window.
+    QuitWorkspaceOrWindow,
+    /// Quit all windows of selected workspace, or container or the selected
+    /// window.
+    QuitWorkspaceContainerOrWindow,
     /// Focus the next window in LRU order.
     NextWindow {
         #[clap(subcommand)]
@@ -102,16 +116,15 @@ pub enum SwayrCommand {
         #[clap(subcommand)]
         windows: ConsiderWindows,
     },
-    /// Quit the selected window.
-    QuitWindow,
-    /// Switch to the selected workspace.
-    SwitchWorkspace,
-    /// Switch to the selected workspace or focus the selected window.
-    SwitchWorkspaceOrWindow,
-    /// Quit all windows of selected workspace or the selected window.
-    QuitWorkspaceOrWindow,
-    /// Move the currently focused window or container to the selected workspace.
+    /// Move the currently focused window or container to the selected
+    /// workspace.
     MoveFocusedToWorkspace,
+    /// Move the currently focused window or container to the selected
+    /// workspace, container or window.
+    MoveFocusedTo,
+    /// Swap the currently focused window or container with the selected
+    /// container or window.
+    SwapFocusedWith,
     /// Tab or shuffle-and-tile the windows on the current workspace, including
     /// or excluding floating windows.
     ToggleTabShuffleTileWorkspace {
@@ -162,17 +175,23 @@ impl SwayrCommand {
 
 pub struct ExecSwayrCmdArgs<'a> {
     pub cmd: &'a SwayrCommand,
-    pub extra_props: Arc<RwLock<HashMap<i64, con::ExtraProps>>>,
+    pub extra_props: Arc<RwLock<HashMap<i64, t::ExtraProps>>>,
 }
 
 impl DisplayFormat for SwayrCommand {
     fn format_for_display(&self, _: &cfg::Config) -> std::string::String {
-        // TODO: Add a format to Config
+        // TODO: It would be very nice if the display format was exactly like
+        // the swayr invocation in the shell.  Can that somehow be retrieved
+        // from clap?
         format!("{:?}", self)
     }
+
+    fn get_indent_level(&self) -> usize {
+        0
+    }
 }
 
-fn always_true(_x: &con::Window) -> bool {
+fn always_true(_x: &t::DisplayNode) -> bool {
     true
 }
 
@@ -200,6 +219,29 @@ pub fn exec_swayr_cmd(args: ExecSwayrCmdArgs) {
             switch_to_urgent_or_lru_window(&*props.read().unwrap())
         }
         SwayrCommand::SwitchWindow => switch_window(&*props.read().unwrap()),
+        SwayrCommand::SwitchWorkspace => {
+            switch_workspace(&*props.read().unwrap())
+        }
+        SwayrCommand::SwitchWorkspaceOrWindow => {
+            switch_workspace_or_window(&*props.read().unwrap())
+        }
+        SwayrCommand::SwitchWorkspaceContainerOrWindow => {
+            switch_workspace_container_or_window(&*props.read().unwrap())
+        }
+        SwayrCommand::QuitWindow => quit_window(&*props.read().unwrap()),
+        SwayrCommand::QuitWorkspaceOrWindow => {
+            quit_workspace_or_window(&*props.read().unwrap())
+        }
+        SwayrCommand::QuitWorkspaceContainerOrWindow => {
+            quit_workspace_container_or_window(&*props.read().unwrap())
+        }
+        SwayrCommand::MoveFocusedToWorkspace => {
+            move_focused_to_workspace(&*props.read().unwrap())
+        }
+        SwayrCommand::MoveFocusedTo => move_focused_to(&*props.read().unwrap()),
+        SwayrCommand::SwapFocusedWith => {
+            swap_focused_with(&*props.read().unwrap())
+        }
         SwayrCommand::NextWindow { windows } => focus_window_in_direction(
             Direction::Forward,
             windows,
@@ -216,16 +258,18 @@ pub fn exec_swayr_cmd(args: ExecSwayrCmdArgs) {
             Direction::Forward,
             windows,
             &*props.read().unwrap(),
-            Box::new(|w: &con::Window| {
-                !w.is_floating() && w.is_child_of_tiled_container()
+            Box::new(|dn: &t::DisplayNode| {
+                !dn.node.is_floating()
+                    && dn.tree.is_child_of_tiled_container(dn.node.id)
             }),
         ),
         SwayrCommand::PrevTiledWindow { windows } => focus_window_in_direction(
             Direction::Backward,
             windows,
             &*props.read().unwrap(),
-            Box::new(|w: &con::Window| {
-                !w.is_floating() && w.is_child_of_tiled_container()
+            Box::new(|dn: &t::DisplayNode| {
+                !dn.node.is_floating()
+                    && dn.tree.is_child_of_tiled_container(dn.node.id)
             }),
         ),
         SwayrCommand::NextTabbedOrStackedWindow { windows } => {
@@ -233,9 +277,11 @@ pub fn exec_swayr_cmd(args: ExecSwayrCmdArgs) {
                 Direction::Forward,
                 windows,
                 &*props.read().unwrap(),
-                Box::new(|w: &con::Window| {
-                    !w.is_floating()
-                        && w.is_child_of_tabbed_or_stacked_container()
+                Box::new(|dn: &t::DisplayNode| {
+                    !dn.node.is_floating()
+                        && dn
+                            .tree
+                            .is_child_of_tabbed_or_stacked_container(dn.node.id)
                 }),
             )
         }
@@ -244,9 +290,11 @@ pub fn exec_swayr_cmd(args: ExecSwayrCmdArgs) {
                 Direction::Backward,
                 windows,
                 &*props.read().unwrap(),
-                Box::new(|w: &con::Window| {
-                    !w.is_floating()
-                        && w.is_child_of_tabbed_or_stacked_container()
+                Box::new(|dn: &t::DisplayNode| {
+                    !dn.node.is_floating()
+                        && dn
+                            .tree
+                            .is_child_of_tabbed_or_stacked_container(dn.node.id)
                 }),
             )
         }
@@ -255,7 +303,7 @@ pub fn exec_swayr_cmd(args: ExecSwayrCmdArgs) {
                 Direction::Forward,
                 windows,
                 &*props.read().unwrap(),
-                Box::new(|w: &con::Window| w.is_floating()),
+                Box::new(|dn: &t::DisplayNode| dn.node.is_floating()),
             )
         }
         SwayrCommand::PrevFloatingWindow { windows } => {
@@ -263,7 +311,7 @@ pub fn exec_swayr_cmd(args: ExecSwayrCmdArgs) {
                 Direction::Backward,
                 windows,
                 &*props.read().unwrap(),
-                Box::new(|w: &con::Window| w.is_floating()),
+                Box::new(|dn: &t::DisplayNode| dn.node.is_floating()),
             )
         }
         SwayrCommand::NextWindowOfSameLayout { windows } => {
@@ -280,19 +328,6 @@ pub fn exec_swayr_cmd(args: ExecSwayrCmdArgs) {
                 &*props.read().unwrap(),
             )
         }
-        SwayrCommand::MoveFocusedToWorkspace => {
-            move_focused_container_to_workspace(&*props.read().unwrap())
-        }
-        SwayrCommand::QuitWindow => quit_window(&*props.read().unwrap()),
-        SwayrCommand::SwitchWorkspace => {
-            switch_workspace(&*props.read().unwrap())
-        }
-        SwayrCommand::SwitchWorkspaceOrWindow => {
-            switch_workspace_or_window(&*props.read().unwrap())
-        }
-        SwayrCommand::QuitWorkspaceOrWindow => {
-            quit_workspace_or_window(&*props.read().unwrap())
-        }
         SwayrCommand::TileWorkspace { floating } => {
             tile_current_workspace(floating, false)
         }
@@ -310,6 +345,7 @@ pub fn exec_swayr_cmd(args: ExecSwayrCmdArgs) {
             let mut cmds = vec![
                 SwayrCommand::ExecuteSwaymsgCommand,
                 SwayrCommand::MoveFocusedToWorkspace,
+                SwayrCommand::SwapFocusedWith,
                 SwayrCommand::QuitWindow,
                 SwayrCommand::QuitWorkspaceOrWindow,
                 SwayrCommand::SwitchWindow,
@@ -375,25 +411,26 @@ fn quit_window_by_id(id: i64) {
     run_sway_command(&[format!("[con_id={}]", id).as_str(), "kill"]);
 }
 
-pub fn get_tree() -> s::Node {
+pub fn get_tree(include_scratchpad: bool) -> s::Node {
     match s::Connection::new() {
-        Ok(mut con) => con.get_tree().expect("Got no root node"),
+        Ok(mut con) => {
+            let mut root = con.get_tree().expect("Got no root node");
+            if !include_scratchpad {
+                root.nodes.retain(|o| !o.is_scratchpad());
+            }
+            root
+        }
         Err(err) => panic!("{}", err),
     }
 }
 
 pub fn switch_to_urgent_or_lru_window(
-    extra_props: &HashMap<i64, con::ExtraProps>,
+    extra_props: &HashMap<i64, t::ExtraProps>,
 ) {
-    let root = get_tree();
-    let windows = con::get_windows(&root, false, extra_props);
-    if let Some(win) = windows
-        .iter()
-        .find(|w| w.is_urgent())
-        .or_else(|| windows.get(0))
-    {
-        println!("Switching to {}, id: {}", win.get_app_name(), win.get_id());
-        focus_window_by_id(win.get_id())
+    let root = get_tree(false);
+    let tree = t::get_tree(&root, extra_props);
+    if let Some(win) = tree.get_windows().get(0) {
+        focus_window_by_id(win.node.id)
     } else {
         println!("No window to switch to.")
     }
@@ -446,146 +483,104 @@ fn handle_non_matching_input(input: &str) {
     }
 }
 
-pub fn switch_window(extra_props: &HashMap<i64, con::ExtraProps>) {
-    let root = get_tree();
-    let windows = con::get_windows(&root, true, extra_props);
-
-    match util::select_from_menu("Switch to window", &windows) {
-        Ok(window) => focus_window_by_id(window.get_id()),
+fn select_and_focus(prompt: &str, choices: &[t::DisplayNode]) {
+    match util::select_from_menu(prompt, choices) {
+        Ok(tn) => match tn.node.get_type() {
+            t::Type::Workspace => {
+                if !tn.node.is_scratchpad() {
+                    run_sway_command(&["workspace", tn.node.get_name()]);
+                }
+            }
+            t::Type::Window | t::Type::Container => {
+                focus_window_by_id(tn.node.id)
+            }
+            t => {
+                eprintln!("Cannot handle {:?} in select_and_focus", t)
+            }
+        },
         Err(non_matching_input) => {
             handle_non_matching_input(&non_matching_input)
         }
     }
 }
 
-pub enum Direction {
-    Backward,
-    Forward,
+pub fn switch_window(extra_props: &HashMap<i64, t::ExtraProps>) {
+    let root = get_tree(true);
+    let tree = t::get_tree(&root, extra_props);
+    select_and_focus("Select window", &tree.get_windows());
 }
 
-pub fn focus_window_in_direction(
-    dir: Direction,
-    consider_wins: &ConsiderWindows,
-    extra_props: &HashMap<i64, con::ExtraProps>,
-    pred: Box<dyn Fn(&con::Window) -> bool>,
-) {
-    let root = get_tree();
-    let root = match consider_wins {
-        ConsiderWindows::AllWorkspaces => &root,
-        ConsiderWindows::CurrentWorkspace => {
-            con::get_workspaces(&root, false, extra_props)
-                .iter()
-                .find(|w| w.is_current())
-                .expect("No current workspace!")
-                .node
-        }
-    };
-
-    let mut windows: Vec<con::Window> =
-        con::get_windows(root, false, extra_props)
-            .into_iter()
-            .filter(|w| pred(w))
-            .collect();
-
-    if windows.len() < 2 {
-        return;
-    }
-
-    windows.sort_by(|a, b| {
-        let lru_a = a.last_focus_time_for_next_prev_seq();
-        let lru_b = b.last_focus_time_for_next_prev_seq();
-        lru_a.cmp(&lru_b).reverse()
-    });
-
-    let is_focused_window: Box<dyn Fn(&con::Window) -> bool> =
-        if !windows.iter().any(|w| w.is_focused()) {
-            let last_focused_win_id = windows.get(0).unwrap().get_id();
-            Box::new(move |w| w.get_id() == last_focused_win_id)
-        } else {
-            Box::new(|w: &con::Window| w.is_focused())
-        };
-
-    let mut iter: Box<dyn Iterator<Item = &con::Window>> = match dir {
-        Direction::Forward => Box::new(windows.iter().rev().cycle()),
-        Direction::Backward => Box::new(windows.iter().cycle()),
-    };
-
-    loop {
-        let win = iter.next().unwrap();
-        if is_focused_window(win) {
-            let win = iter.next().unwrap();
-            focus_window_by_id(win.get_id());
-            return;
-        }
-    }
+pub fn switch_workspace(extra_props: &HashMap<i64, t::ExtraProps>) {
+    let root = get_tree(false);
+    let tree = t::get_tree(&root, extra_props);
+    select_and_focus("Select workspace", &tree.get_workspaces());
 }
 
-pub fn focus_window_of_same_layout_in_direction(
-    dir: Direction,
-    consider_wins: &ConsiderWindows,
-    extra_props: &HashMap<i64, con::ExtraProps>,
-) {
-    let root = get_tree();
-    let windows = con::get_windows(&root, false, extra_props);
-    let current_window = windows
-        .iter()
-        .find(|w| w.is_focused())
-        .or_else(|| windows.get(0));
-
-    if let Some(current_window) = current_window {
-        focus_window_in_direction(
-            dir,
-            consider_wins,
-            extra_props,
-            if current_window.is_floating() {
-                Box::new(|w| w.is_floating())
-            } else if !current_window.is_floating()
-                && current_window.is_child_of_tabbed_or_stacked_container()
-            {
-                Box::new(|w| {
-                    !w.is_floating()
-                        && w.is_child_of_tabbed_or_stacked_container()
-                })
-            } else if !current_window.is_floating()
-                && current_window.is_child_of_tiled_container()
-            {
-                Box::new(|w| {
-                    !w.is_floating() && w.is_child_of_tiled_container()
-                })
-            } else {
-                Box::new(always_true)
-            },
-        )
-    }
-}
-
-pub fn switch_workspace(extra_props: &HashMap<i64, con::ExtraProps>) {
-    let root = get_tree();
-    let workspaces = con::get_workspaces(&root, false, extra_props);
-
-    match util::select_from_menu("Switch to workspace", &workspaces) {
-        Ok(workspace) => run_sway_command(&["workspace", workspace.get_name()]),
-        Err(non_matching_input) => {
-            handle_non_matching_input(&non_matching_input)
-        }
-    }
-}
-
-pub fn move_focused_container_to_workspace(
-    extra_props: &HashMap<i64, con::ExtraProps>,
-) {
-    let root = get_tree();
-    let workspaces = con::get_workspaces(&root, false, extra_props);
-
-    let val = util::select_from_menu(
-        "Move focused container to workspace",
-        &workspaces,
+pub fn switch_workspace_or_window(extra_props: &HashMap<i64, t::ExtraProps>) {
+    let root = get_tree(true);
+    let tree = t::get_tree(&root, extra_props);
+    select_and_focus(
+        "Select workspace or window",
+        &tree.get_workspaces_and_windows(),
     );
-    let ws_name = &match val {
-        Ok(workspace) => String::from(workspace.get_name()),
-        Err(input) => String::from(chop_workspace_shortcut(&input)),
-    };
+}
 
+pub fn switch_workspace_container_or_window(
+    extra_props: &HashMap<i64, t::ExtraProps>,
+) {
+    let root = get_tree(true);
+    let tree = t::get_tree(&root, extra_props);
+    select_and_focus(
+        "Select workspace, container or window",
+        &tree.get_workspaces_containers_and_windows(),
+    );
+}
+
+fn select_and_quit(prompt: &str, choices: &[t::DisplayNode]) {
+    if let Ok(tn) = util::select_from_menu(prompt, choices) {
+        match tn.node.get_type() {
+            t::Type::Workspace | t::Type::Container => {
+                for win in
+                    tn.node.iter().filter(|n| n.get_type() == t::Type::Window)
+                {
+                    quit_window_by_id(win.id)
+                }
+            }
+            t::Type::Window => quit_window_by_id(tn.node.id),
+            t => {
+                eprintln!("Cannot handle {:?} in quit_workspace_or_window", t)
+            }
+        }
+    }
+}
+
+pub fn quit_window(extra_props: &HashMap<i64, t::ExtraProps>) {
+    let root = get_tree(true);
+    let tree = t::get_tree(&root, extra_props);
+    select_and_quit("Quit window", &tree.get_windows());
+}
+
+pub fn quit_workspace_or_window(extra_props: &HashMap<i64, t::ExtraProps>) {
+    let root = get_tree(true);
+    let tree = t::get_tree(&root, extra_props);
+    select_and_quit(
+        "Quit workspace or window",
+        &tree.get_workspaces_and_windows(),
+    );
+}
+
+pub fn quit_workspace_container_or_window(
+    extra_props: &HashMap<i64, t::ExtraProps>,
+) {
+    let root = get_tree(true);
+    let tree = t::get_tree(&root, extra_props);
+    select_and_quit(
+        "Quit workspace, container or window",
+        &tree.get_workspaces_containers_and_windows(),
+    );
+}
+
+fn move_focused_to_workspace_1(ws_name: &str) {
     if DIGIT_AND_NAME.is_match(ws_name) {
         run_sway_command(&[
             "move",
@@ -600,50 +595,182 @@ pub fn move_focused_container_to_workspace(
     }
 }
 
-pub fn switch_workspace_or_window(extra_props: &HashMap<i64, con::ExtraProps>) {
-    let root = get_tree();
-    let workspaces = con::get_workspaces(&root, true, extra_props);
-    let ws_or_wins = con::WsOrWin::from_workspaces(&workspaces);
-    match util::select_from_menu("Select workspace or window", &ws_or_wins) {
-        Ok(ws_or_win) => match ws_or_win {
-            con::WsOrWin::Ws { ws } => {
-                run_sway_command(&["workspace", ws.get_name()]);
-            }
-            con::WsOrWin::Win { win } => focus_window_by_id(win.get_id()),
-        },
-        Err(non_matching_input) => {
-            handle_non_matching_input(&non_matching_input)
-        }
-    }
+fn move_focused_to_container_or_window(id: i64) {
+    run_sway_command(&[
+        &format!("[con_id=\"{}\"]", id),
+        "mark",
+        "--add",
+        "__SWAYR_MOVE_TARGET__",
+    ]);
+    run_sway_command(&["move", "to", "mark", "__SWAYR_MOVE_TARGET__"]);
+    run_sway_command(&["unmark", "__SWAYR_MOVE_TARGET__"]);
 }
 
-pub fn quit_window(extra_props: &HashMap<i64, con::ExtraProps>) {
-    let root = get_tree();
-    let windows = con::get_windows(&root, true, extra_props);
-
-    if let Ok(window) = util::select_from_menu("Quit window", &windows) {
-        quit_window_by_id(window.get_id())
-    }
-}
-
-pub fn quit_workspace_or_window(extra_props: &HashMap<i64, con::ExtraProps>) {
-    let root = get_tree();
-    let workspaces = con::get_workspaces(&root, true, extra_props);
-    let ws_or_wins = con::WsOrWin::from_workspaces(&workspaces);
-    if let Ok(ws_or_win) =
-        util::select_from_menu("Quit workspace or window", &ws_or_wins)
-    {
-        match ws_or_win {
-            con::WsOrWin::Ws { ws } => {
-                for win in ws.get_windows() {
-                    quit_window_by_id(win.get_id())
+fn select_and_move_focused_to(prompt: &str, choices: &[t::DisplayNode]) {
+    match util::select_from_menu(prompt, choices) {
+        Ok(tn) => match tn.node.get_type() {
+            t::Type::Workspace => {
+                if tn.node.is_scratchpad() {
+                    run_sway_command(&["move", "container", "to", "scratchpad"])
+                } else {
+                    move_focused_to_workspace_1(tn.node.get_name())
                 }
             }
-            con::WsOrWin::Win { win } => quit_window_by_id(win.get_id()),
+            t::Type::Container | t::Type::Window => {
+                move_focused_to_container_or_window(tn.node.id)
+            }
+            t => eprintln!("Cannot move focused to {:?}", t),
+        },
+        Err(input) => {
+            let ws_name = chop_workspace_shortcut(&input);
+            move_focused_to_workspace_1(ws_name);
         }
     }
 }
 
+pub fn move_focused_to_workspace(extra_props: &HashMap<i64, t::ExtraProps>) {
+    let root = get_tree(true);
+    let tree = t::get_tree(&root, extra_props);
+    select_and_move_focused_to(
+        "Move focused container to workspace",
+        &tree.get_workspaces(),
+    );
+}
+
+pub fn move_focused_to(extra_props: &HashMap<i64, t::ExtraProps>) {
+    let root = get_tree(true);
+    let tree = t::get_tree(&root, extra_props);
+    select_and_move_focused_to(
+        "Move focused container to workspace or container",
+        &tree.get_workspaces_containers_and_windows(),
+    );
+}
+
+pub fn swap_focused_with(extra_props: &HashMap<i64, t::ExtraProps>) {
+    let root = get_tree(true);
+    let tree = t::get_tree(&root, extra_props);
+    match util::select_from_menu(
+        "Swap focused with",
+        &tree.get_workspaces_containers_and_windows(),
+    ) {
+        Ok(tn) => match tn.node.get_type() {
+            t::Type::Workspace | t::Type::Container | t::Type::Window => {
+                run_sway_command(&[
+                    "swap",
+                    "container",
+                    "with",
+                    "con_id",
+                    &format!("{}", tn.node.id),
+                ])
+            }
+            t => eprintln!("Cannot move focused to {:?}", t),
+        },
+        Err(input) => {
+            let ws_name = chop_workspace_shortcut(&input);
+            move_focused_to_workspace_1(ws_name);
+        }
+    }
+}
+
+pub enum Direction {
+    Backward,
+    Forward,
+}
+
+pub fn focus_window_in_direction(
+    dir: Direction,
+    consider_wins: &ConsiderWindows,
+    extra_props: &HashMap<i64, t::ExtraProps>,
+    pred: Box<dyn Fn(&t::DisplayNode) -> bool>,
+) {
+    let root = get_tree(false);
+    let tree = t::get_tree(&root, extra_props);
+    let mut wins = tree.get_windows();
+
+    if consider_wins == &ConsiderWindows::CurrentWorkspace {
+        let cur_ws = tree.get_current_workspace();
+        wins.retain(|w| {
+            tree.get_workspace_node(w.node.id).unwrap().id == cur_ws.id
+        });
+    }
+
+    wins.retain(pred);
+
+    if wins.len() < 2 {
+        return;
+    }
+
+    wins.sort_by(|a, b| {
+        let lru_a = tree.last_focus_time_for_next_prev_seq(a.node.id);
+        let lru_b = tree.last_focus_time_for_next_prev_seq(b.node.id);
+        lru_a.cmp(&lru_b).reverse()
+    });
+
+    let is_focused_window: Box<dyn Fn(&t::DisplayNode) -> bool> =
+        if !wins.iter().any(|w| w.node.focused) {
+            let last_focused_win_id = wins.get(0).unwrap().node.id;
+            Box::new(move |dn| dn.node.id == last_focused_win_id)
+        } else {
+            Box::new(|dn| dn.node.focused)
+        };
+
+    let mut iter: Box<dyn Iterator<Item = &t::DisplayNode>> = match dir {
+        Direction::Forward => Box::new(wins.iter().rev().cycle()),
+        Direction::Backward => Box::new(wins.iter().cycle()),
+    };
+
+    loop {
+        let win = iter.next().unwrap();
+        if is_focused_window(win) {
+            let win = iter.next().unwrap();
+            focus_window_by_id(win.node.id);
+            return;
+        }
+    }
+}
+
+pub fn focus_window_of_same_layout_in_direction(
+    dir: Direction,
+    consider_wins: &ConsiderWindows,
+    extra_props: &HashMap<i64, t::ExtraProps>,
+) {
+    let root = get_tree(false);
+    let tree = t::get_tree(&root, extra_props);
+    let wins = tree.get_windows();
+    let cur_win = wins.get(0);
+
+    if let Some(cur_win) = cur_win {
+        focus_window_in_direction(
+            dir,
+            consider_wins,
+            extra_props,
+            if cur_win.node.is_floating() {
+                Box::new(|dn| dn.node.is_floating())
+            } else if !cur_win.node.is_floating()
+                && cur_win
+                    .tree
+                    .is_child_of_tabbed_or_stacked_container(cur_win.node.id)
+            {
+                Box::new(|dn| {
+                    !dn.node.is_floating()
+                        && dn
+                            .tree
+                            .is_child_of_tabbed_or_stacked_container(dn.node.id)
+                })
+            } else if !cur_win.node.is_floating()
+                && cur_win.tree.is_child_of_tiled_container(cur_win.node.id)
+            {
+                Box::new(|dn| {
+                    !dn.node.is_floating()
+                        && dn.tree.is_child_of_tiled_container(dn.node.id)
+                })
+            } else {
+                Box::new(always_true)
+            },
+        )
+    }
+}
+
 fn tile_current_workspace(floating: &ConsiderFloating, shuffle: bool) {
     match layout::relayout_current_workspace(
         floating == &ConsiderFloating::IncludeFloating,
@@ -662,22 +789,19 @@ fn tile_current_workspace(floating: &ConsiderFloating, shuffle: bool) {
                 if win.is_floating() {
                     con.run_command(format!(
                         "[con_id={}] floating disable",
-                        win.get_id()
+                        win.id
                     ))?;
                 }
                 std::thread::sleep(std::time::Duration::from_millis(25));
                 con.run_command(format!(
                     "[con_id={}] move to workspace current",
-                    win.get_id()
+                    win.id
                 ))?;
                 placed_wins.push(win);
                 if shuffle {
                     std::thread::sleep(std::time::Duration::from_millis(25));
                     if let Some(win) = placed_wins.choose(&mut rng) {
-                        con.run_command(format!(
-                            "[con_id={}] focus",
-                            win.get_id()
-                        ))?;
+                        con.run_command(format!("[con_id={}] focus", win.id))?;
                     }
                 }
             }
@@ -702,14 +826,14 @@ fn tab_current_workspace(floating: &ConsiderFloating) {
                 if win.is_floating() {
                     con.run_command(format!(
                         "[con_id={}] floating disable",
-                        win.get_id()
+                        win.id
                     ))?;
                 }
 
                 std::thread::sleep(std::time::Duration::from_millis(25));
                 con.run_command(format!(
                     "[con_id={}] move to workspace current",
-                    win.get_id()
+                    win.id
                 ))?;
                 placed_wins.push(win);
             }
@@ -722,12 +846,9 @@ fn tab_current_workspace(floating: &ConsiderFloating) {
 }
 
 fn toggle_tab_tile_current_workspace(floating: &ConsiderFloating) {
-    let tree = get_tree();
-    let workspaces = tree.workspaces();
-    let cur_ws = workspaces
-        .iter()
-        .find(|w| con::is_current_container(w))
-        .unwrap();
+    let tree = get_tree(false);
+    let workspaces = tree.nodes_of_type(t::Type::Workspace);
+    let cur_ws = workspaces.iter().find(|w| w.is_current()).unwrap();
     if cur_ws.layout == s::NodeLayout::Tabbed {
         tile_current_workspace(floating, true);
     } else {
@@ -839,6 +960,10 @@ impl DisplayFormat for SwaymsgCmd<'_> {
     fn format_for_display(&self, _: &cfg::Config) -> std::string::String {
         self.cmd.join(" ")
     }
+
+    fn get_indent_level(&self) -> usize {
+        0
+    }
 }
 
 pub fn exec_swaymsg_command() {
